/* hash.c - IRC network state database
 * Copyright 2000-2001 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: hash.c,v 1.121.2.19 2001/11/28 02:09:43 entrope Exp $
 */

#include "config.h"
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#include "common.h"
#include "commands.h"
#include "dict.h"
#include "conf.h"
#include "hash.h"
#include "global.h"
#include "log.h"
#include "nickserv.h"
#include "chanserv.h"
#include "policer.h"
#include "parse.h"
#include "recdb.h"

extern time_t boot_time;
extern int burst;
extern struct policer_params *oper_policer_params, *luser_policer_params;

struct server *self;
struct server *servers[64];
dict_t channels;
dict_t clients;
unsigned int max_clients, invis_clients;
time_t max_clients_time;
unsigned int server_count = 0;
struct userList curr_opers;

static unsigned int next_numeric = 0;
unsigned int highest_numeric = 0;

void init_structs(void)
{
	unsigned int i;
	char numer[COMBO_NUMERIC_LEN+1];

	for (i=0; i<ArrayLength(servers); i++) {
	    servers[i] = NULL;
	}

	channels = dict_new();
	clients = dict_new();
	userList_init(&curr_opers);

        reg_exit_func(local_split);

	inttobase64(numer, conf_server_num(), SERVER_NUMERIC_LEN);
	for (i=SERVER_NUMERIC_LEN; i<COMBO_NUMERIC_LEN+1; i++) numer[i] = ']';
	numer[COMBO_NUMERIC_LEN] = 0;
	self = AddServer(NULL, conf_server_name(), 0, boot_time, now, numer, conf_server_desc());
}

void local_split(void)
{
    /* remove all but local server users and juped/virtual servers, etc */
    unsigned int i;
    for (i=0;i<ArrayLength(servers);i++) {
	if (servers[i] && (servers[i] != self)) {
	    DelServer(servers[i], false, "");
	    servers[i]=NULL;
	}
    }
}

static int
foreach_burst_helper(const char *key, void *data, void *extra)
{
    (void)key; (void)extra;
    irc_burst(data);
    return 0;
}

void send_burst(void)
{
	unsigned int i, hop, max_hop=1;
	
	/* burst servers, closest first (don't need to burst this server, it was sent during link auth) */
	for (i=0; i<ArrayLength(servers); i++) {
	    if (servers[i] && servers[i]->hops > max_hop) {
		max_hop = servers[i]->hops;
	    }
	}
	for (hop=1;hop<=max_hop;hop++) {
	    for (i=0;i<ArrayLength(servers);i++) {
		if (!servers[i]) continue;

		if (servers[i]->hops == hop)
		    irc_server(servers[i]);
	    }
	}

	/* burst local nicks */
	for (i=0; i<highest_numeric; i++) {
	    irc_user(self->users[i]);
	}

	/* burst channels */
	dict_foreach(channels, foreach_burst_helper, NULL);

	irc_eob();
}

static server_link_func_t *slf_list;
static unsigned int slf_size = 0, slf_used = 0;

void
reg_server_link_func(server_link_func_t handler)
{
    if (slf_used == slf_size) {
	if (slf_size) {
	    slf_size <<= 1;
	    slf_list = realloc(slf_list, slf_size*sizeof(server_link_func_t));
	} else {
	    slf_size = 8;
	    slf_list = malloc(slf_size*sizeof(server_link_func_t));
	}
    }
    slf_list[slf_used++] = handler;
}

struct server *
AddServer(struct server *uplink, const char *name, int hops, time_t boot, time_t link, const char *numeric, const char *description)
{
	struct server* sNode;

	if ((sNode = GetServer(numeric))) {
            /* This means we're trying to re-add an existant server.
             * To be safe, we should forget the previous incarnation.
             * (And all its linked servers.)
             *
             * It usually only happens in replays when the original
             * had a ping timeout and the replay didn't (because
             * replaying a ping timeout invariably gets things wrong).
             */
            DelServer(sNode, 0, NULL);
        }

        sNode = malloc(sizeof(server));
        servers[base64toint(numeric, SERVER_NUMERIC_LEN)] = sNode;
	sNode->uplink = uplink;
	safestrncpy(sNode->name, name, sizeof(sNode->name));
        sNode->num_mask = base64toint(numeric+SERVER_NUMERIC_LEN, USER_NUMERIC_LEN);
	sNode->hops = hops;
	sNode->clients = 0;
	sNode->boot = boot;
	sNode->link = link;
        safestrncpy(sNode->numeric, numeric, sizeof(sNode->numeric));
	safestrncpy(sNode->description, description, sizeof(sNode->description));
        sNode->users = malloc((sNode->num_mask+1) * sizeof(*sNode->users));
        memset(sNode->users, 0, (sNode->num_mask+1) * sizeof(*sNode->users));
	
	if (IsLocal(numeric)) irc_server(sNode);

	if (hops && !burst) {
	    unsigned int n;
	    for (n=0; n<slf_used; n++) {
		slf_list[n](sNode);
	    }
	}

	server_count++;
	return sNode;
}

void DelServer(server* serv, int announce, const char *message)
{
	unsigned int i;

	/* If we receive an ERROR command before the SERVER
	 * command a NULL server can be passed */
	if (!serv) return;

	if (announce && IsLocal(serv->numeric)) /* send squit message */
	    irc_squit(serv, message);

	/* must recursively remove servers linked to this one first */
	for (i=0;i<ArrayLength(servers);i++) {
	    if (servers[i] && servers[i]->uplink == serv) {
		if(!servers[i]->uplink || servers[i] == self) break;
		DelServer(servers[i], false, "");
	    }
	}

	/* clean up server's user hash tables */
	for (i=0;i<serv->num_mask;i++) {
	    if (serv->users[i]) {
		DelUser(serv->users[i], NULL, false, "server delinked");
	    }
	}

	/* delete server */
	servers[base64toint(serv->numeric, 1)]=NULL;
	free(serv);
	server_count--;
}

server* GetServer(const char *numeric)
{
    return (numeric[0]) ? servers[base64toint(numeric, SERVER_NUMERIC_LEN)] : NULL;
}

server*
GetServerH(const char *name)
{
    unsigned int n;
    for (n=0; n<ArrayLength(servers); n++) {
	if (!servers[n]) continue;
	if (!strcasecmp(name, servers[n]->name)) return servers[n];
    }
    return NULL;
}

static int
get_local_numeric(void)
{
    unsigned int last = next_numeric;
    do {
	if (!self->users[next_numeric++]) {
	    if(next_numeric > highest_numeric) {
		highest_numeric = next_numeric;
	    }

	    return next_numeric-1;
	}
	if (next_numeric == self->num_mask) next_numeric = 0;
    } while (next_numeric != last);
    return -1;
}

struct userNode *
AddService(const char *nick, const char *desc)
{
    char numeric[COMBO_NUMERIC_LEN+1], ip_num[7];
    int local_num = get_local_numeric();

    if (local_num == -1) return 0;
    safestrncpy(numeric, self->numeric, SERVER_NUMERIC_LEN+1);
    inttobase64(numeric+SERVER_NUMERIC_LEN, local_num, USER_NUMERIC_LEN);
    numeric[COMBO_NUMERIC_LEN] = 0;
    inttobase64(ip_num, 127 << 24 | 1, 6);
    return AddUser(self, nick, nick, self->name, "+oik", numeric, desc, 0, ip_num);
}

struct userNode *
AddClone(const char *nick, const char *username, const char *hostname, const char *desc)
{
    char numeric[COMBO_NUMERIC_LEN+1], ip_num[7];
    int local_num = get_local_numeric();

    if (local_num == -1) return 0;
    safestrncpy(numeric, self->numeric, SERVER_NUMERIC_LEN+1);
    inttobase64(numeric+SERVER_NUMERIC_LEN, local_num, USER_NUMERIC_LEN);
    numeric[COMBO_NUMERIC_LEN] = 0;
    inttobase64(ip_num, 127 << 24 | 1, 6);
    return AddUser(self, nick, username, hostname, "+i", numeric, desc, 0, ip_num);
}

static new_user_func_t *nuf_list;
static unsigned int nuf_size = 0, nuf_used = 0;

void
reg_new_user_func(new_user_func_t handler)
{
    if (nuf_used == nuf_size) {
	if (nuf_size) {
	    nuf_size <<= 1;
	    nuf_list = realloc(nuf_list, nuf_size*sizeof(new_user_func_t));
	} else {
	    nuf_size = 8;
	    nuf_list = malloc(nuf_size*sizeof(new_user_func_t));
	}
    }
    nuf_list[nuf_used++] = handler;
}

static nick_change_func_t *ncf2_list;
static unsigned int ncf2_size = 0, ncf2_used = 0;

void
reg_nick_change_func(nick_change_func_t handler)
{
    if (ncf2_used == ncf2_size) {
        if (ncf2_size) {
            ncf2_size <<= 1;
            ncf2_list = realloc(ncf2_list, ncf2_size*sizeof(nick_change_func_t));
        } else {
            ncf2_size = 8;
            ncf2_list = malloc(ncf2_size*sizeof(nick_change_func_t));
        }
    }
    ncf2_list[ncf2_used++] = handler;
}


static del_user_func_t *duf_list;
static unsigned int duf_size = 0, duf_used = 0;

void
reg_del_user_func(del_user_func_t handler)
{
    if (duf_used == duf_size) {
	if (duf_size) {
	    duf_size <<= 1;
	    duf_list = realloc(duf_list, duf_size*sizeof(del_user_func_t));
	} else {
	    duf_size = 8;
	    duf_list = malloc(duf_size*sizeof(del_user_func_t));
	}
    }
    duf_list[duf_used++] = handler;
}

void
unreg_del_user_func(del_user_func_t handler)
{
    unsigned int i;
    for (i=0; i<duf_used; i++) {
        if (duf_list[i] == handler) break;
    }
    if (i == duf_used) return;
    memmove(duf_list+i, duf_list+i+1, (duf_used-i-1)*sizeof(duf_list[0]));
    duf_used--;
}

userNode* AddUser(server* uplink, const char *nick, const char *ident, const char *hostname, const char *modes, const char *numeric, const char *userinfo, time_t timestamp, const char *realip)
{
	userNode* uNode;
	struct policer_params *pol_params;
	unsigned int n;

	if (strlen(numeric)!=COMBO_NUMERIC_LEN) {
		log(DEBUG_LOG, LOG_WARNING, "AddUser(%p, %s, ...): numeric %s wrong length!\n", uplink, nick, numeric);
		return NULL;
	}

	if (!uplink || !GetServer(numeric)) {
		log(DEBUG_LOG, LOG_WARNING, "AddUser(%p, %s, ...): server for numeric %s doesn't exist!\n", uplink, nick, numeric);
		return NULL;
	}

	if ((uNode = GetUserH(nick))) {
	    if (uNode->uplink == self) {
		char kill[512];
		/* Since we want services to be authoritative over
		 * other servers, we'll issue a KILL regardless of
		 * timestamp comparisons.  If this other user is +k,
		 * then somebody screwed up.  */
		sprintf(kill, "%s %s %s :%s (Nick collision.)", self->numeric, CMD_KILL, numeric, self->name);
                irc_raw(kill);
		return NULL;
	    } else {
		/* Assume the other server knows what's going on. */
		DelUser(uNode, NULL, false, NULL);
	    }
	}

	/* create new usernode and set all values */
	uNode = malloc(sizeof(userNode));
	safestrncpy(uNode->nick, nick, sizeof(uNode->nick));
	safestrncpy(uNode->username, ident, sizeof(uNode->username));
	safestrncpy(uNode->info, userinfo, sizeof(uNode->info));
	safestrncpy(uNode->hostname, hostname, sizeof(uNode->hostname));
	safestrncpy(uNode->numeric, numeric, sizeof(uNode->numeric));
	uNode->ip = base64toint(realip, 6);
	uNode->timestamp=timestamp;
	modeList_init(&uNode->channels);
	uNode->uplink = uplink;
	uNode->uplink->clients++;
	uNode->modes = 0;
	uNode->dead = 0;
	uNode->command_policer = NULL;
	mod_usermode(uNode, modes);
	uNode->handle_info = NULL;
	uNode->next_authed = NULL;
	pol_params = IsOper(uNode) ? oper_policer_params : luser_policer_params;
	uNode->command_policer = policer_new(pol_params, uNode->timestamp);
	uNode->num_local = base64toint(numeric+SERVER_NUMERIC_LEN, USER_NUMERIC_LEN) & uNode->uplink->num_mask;

	/* insert into global nick dictionary */
	dict_insert(clients, uNode->nick, uNode);
	if (dict_size(clients) > max_clients) {
	    max_clients = dict_size(clients);
	    max_clients_time = now;
	}

	GetServer(numeric)->users[uNode->num_local]=uNode;

	if (IsLocal(numeric)) irc_user(uNode);

	if ((uNode->uplink != self) && !burst) {
	    for (n=0; n<nuf_used; n++) {
		if (nuf_list[n](uNode)) break;
	    }
	}

	return uNode;
}

/* removes user from it's server's hash table and nick hash table */
void DelUser(userNode* user, userNode *killer, int announce, const char *why)
{
	unsigned int n;

	/* Call these in reverse order so ChanServ can update presence
	   information before NickServ nukes the handle_info. */
	for (n = duf_used; n > 0; ) duf_list[--n](user);

	user->uplink->clients--;
	user->uplink->users[user->num_local] = NULL;
	if (IsOper(user)) userList_remove(&curr_opers, user);
	/* remove from global dictionary */
	dict_remove(clients, user->nick);

        if (IsInvisible(user)) invis_clients--;

	if (announce) {
	    if (IsLocal(user->numeric)) {
		irc_quit(user, why);
	    } else {
		irc_kill(killer, user, why);
	    }
	}

	/* remove user from all channels */
	for (n=user->channels.used; n>0;) {
	    DelChannelUser(user, user->channels.list[--n]->channel, false, 0);
	}
        modeList_clean(&user->channels);
        policer_delete(user->command_policer);
        /* We don't free them, in case we try to privmsg them or something
         * (like when a stupid oper kills themself).  We just put them onto
         * a list of clients that get freed after processing each line.
         */
	user->dead = 1;
        userList_append(&dead_users, user);
}

/* reintroduces a user after it has been killed. */
void ReintroduceUser(struct userNode *user)
{	
    unsigned int n;
	
    irc_user(user);

    /* rejoin all channels. */
    for (n = 0; n < user->channels.used; n++) {
        struct modeNode *mn = user->channels.list[n];
        char modes[16];
	irc_join(user, mn->channel);
        switch (mn->modes & (MODE_CHANOP | MODE_VOICE)) {
        case 0: modes[0] = 0; break;
        case MODE_VOICE: snprintf(modes, sizeof(modes), "+v %s", user->numeric); break;
        case MODE_CHANOP: snprintf(modes, sizeof(modes), "+o %s", user->numeric); break;
        case MODE_CHANOP|MODE_VOICE: snprintf(modes, sizeof(modes), "+ov %s %s", user->numeric, user->numeric); break;
        }
        if (modes[0]) irc_mode(self->numeric, mn->channel, modes);
    }
}

void NickChange(userNode* user, const char *new_nick)
{
    unsigned int nn;

    /* don't do anything if there's no change */
    if (!strncmp(new_nick, user->nick, NICKLEN)) return;

    /* remove old entry from clients dictionary */
    dict_remove(clients, user->nick);

    /* Make callbacks for nick changes.
     * This needs to be done with the old nick still in place, else
     * dict lookups based on user->nick will miss the user.
     * It is safe to re-insert using new_nick (which may be a
     * temporary pointer) since the later value of user->nick will be
     * correct. */
    for (nn=0; nn<ncf2_used; nn++) {
        ncf2_list[nn](user, new_nick);
    }

    /* copy new nick into struct */
    safestrncpy(user->nick, new_nick, sizeof(user->nick));

    /* reinsert link according to new nick */
    dict_insert(clients, user->nick, user);

    if (IsLocal(user->numeric)) irc_nick(user);

}

userNode* GetUserN(const char *numeric) /* using numeric */
{
    struct userNode *un;
    struct server *s;
    int n;

    if (strlen(numeric)<COMBO_NUMERIC_LEN) {
	/* don't check for more because this could be used during
         * bursts with channel modes following the numeric it */
	log(DEBUG_LOG, LOG_WARNING, "GetUserN(%s) numeric too short!\n", numeric);
	return NULL;
    }
    if (!(s = GetServer(numeric))) {
	log(DEBUG_LOG, LOG_WARNING, "GetUserN(%s) couldn't find server!\n", numeric);
	return NULL;
    }
    n = base64toint(numeric+SERVER_NUMERIC_LEN, USER_NUMERIC_LEN) & s->num_mask;
    if (!(un = s->users[n])) {
	log(DEBUG_LOG, LOG_WARNING, "GetUserN(%s) couldn't find user!\n", numeric);
    }
    return un;
}

struct userNode *
GetUserH(const char *nick)
{
    return dict_find(clients, nick, NULL);
}

struct userNode *
GetUser(const char *user)
{
    return (user[0] == ':') ? GetUserH(user+1) : GetUserN(user);
}

static new_channel_func_t *ncf_list;
static unsigned int ncf_size = 0, ncf_used = 0;

void
reg_new_channel_func(new_channel_func_t handler)
{
    if (ncf_used == ncf_size) {
	if (ncf_size) {
	    ncf_size <<= 1;
	    ncf_list = realloc(ncf_list, ncf_size*sizeof(ncf_list[0]));
	} else {
	    ncf_size = 8;
	    ncf_list = malloc(ncf_size*sizeof(ncf_list[0]));
	}
    }
    ncf_list[ncf_used++] = handler;
}

static void
BurstChannel(struct chanNode *chan, char *members, char *banlist, int call_ncf)
{
    unsigned int i;
    long mode;
    userNode *un;
    modeNode *mn;

    if (call_ncf) {
	for (i=0; i<ncf_used; i++) {
	    ncf_list[i](chan);
	}
    }

    /* go through the list of users to add and add each one */
    for (mode=0, i=0; members[i]; ) {
	char *user = members + i;
	char *end = user + COMBO_NUMERIC_LEN;
	if (!*user) break;
	if (*end == ':') {
	    mode = 0;
	    while (1) {
		end++;
		if (*end == 'o') mode |= MODE_CHANOP;
		else if (*end == 'v') mode |= MODE_VOICE;
		else break;
	    }
	}
	if (*end == ',') end++;
	i = end - members;
	un = GetUserN(user);
	if (!un) continue;
	mn = AddChannelUser(un, chan, 1);
	mn->modes = mode;
    }

    if (banlist) {
	for (i=0; banlist[i];) {
	    char *ban = banlist+i;
	    while (banlist[i] != ' ' && banlist[i]) i++;
	    while (banlist[i] == ' ') banlist[i++]='\0';
	    AddChannelBan(1, &ban, chan, NULL, now, 0);
	}
    }
}

static join_func_t *jf_list;
static unsigned int jf_size = 0, jf_used = 0;

void
reg_join_func(join_func_t handler)
{
    if (jf_used == jf_size) {
	if (jf_size) {
	    jf_size <<= 1;
	    jf_list = realloc(jf_list, jf_size*sizeof(join_func_t));
	} else {
	    jf_size = 8;
	    jf_list = malloc(jf_size*sizeof(join_func_t));
	}
    }
    jf_list[jf_used++] = handler;
}

chanNode* AddChannel(const char *name, time_t time_, const char *modes, const char *key, int limit, char *members, char *banlist)
{
	chanNode* cNode;
	int new;
	char *reason, *message;
	unsigned int n;

	if (!IsChannelName(name)) {
	    log(MAIN_LOG, LOG_ERROR, "Somebody asked to add channel '%s', which isn't a channel name!\n", name);
	    return NULL;
	}

	/* Don't warn while we're bursting.  But we still need to check for
         * members being empty because other servers can burst to us later,
         * and we may just get a banlist on some of the burst lines.
         */
	if (opserv_chan_warn && !burst && members[0]
            && !IsOper(GetUserN(members))
            && (reason = dict_find(opserv_chan_warn, name, NULL)))
        {
	    message = alloca(strlen(reason) + strlen(name) + 55);
	    sprintf(message, "Channel activity warning for channel '%s' (%s).", name, reason);
	    global_message(MESSAGE_RECIPIENT_OPERS, message);
	}

	cNode = GetChannel(name);
	if (!cNode) {
	    new = 1;
	    cNode = malloc(sizeof(chanNode));
            memset(cNode, 0, sizeof(*cNode));
	    safestrncpy(cNode->name, name, sizeof(cNode->name));
	    banList_init(&cNode->banlist);
	    modeList_init(&cNode->members);
	} else {
	    new = 0;
	}

	mod_chanmode(cNode, modes, (char *)key, limit);

	if (new || (cNode->timestamp == 0) || (cNode->timestamp > time_)) {
	    if (key) {
		safestrncpy(cNode->key, key, sizeof(cNode->key));
	    } else {
		cNode->key[0] = 0;
	    }
	    cNode->limit = limit;
	    cNode->timestamp = time_;

	    /* remove our old ban list, replace it with ours */
	    for (n=0; n<cNode->banlist.used; n++) {
		free(cNode->banlist.list[n]);
	    }
	    cNode->banlist.used = 0;
	} else {
	    /* The 'banlist' was originally NULL, which i'm not sure is
	       correct thing to do - do we really want to ignore newer
	       bans?

	       TODO: tell other server that we have the right ban list */
	}
	if (new) dict_insert(channels, cNode->name, cNode);
	BurstChannel(cNode, members, banlist, burst || new);

	return cNode;
}

static del_channel_func_t *dcf_list;
static unsigned int dcf_size = 0, dcf_used = 0;

void
reg_del_channel_func(del_channel_func_t handler)
{
    if (dcf_used == dcf_size) {
	if (dcf_size) {
	    dcf_size <<= 1;
	    dcf_list = realloc(dcf_list, dcf_size*sizeof(dcf_list[0]));
	} else {
	    dcf_size = 8;
	    dcf_list = malloc(dcf_size*sizeof(dcf_list[0]));
	}
    }
    dcf_list[dcf_used++] = handler;
}

void DelChannel(chanNode* channel)
{
    unsigned int n;
    char *argh;

    dict_remove(channels, channel->name);

    if (channel->members.used) {
        log(MAIN_LOG, LOG_ERROR, "Warning: deleting channel %s with %d users remaining\n", channel->name, channel->members.used);
    }

    /* go through all channel members and delete them from the channel */
    for (n=channel->members.used; n>0; ) {
	DelChannelUser(channel->members.list[--n]->user, channel, false, 1);
    }

    /* delete all channel bans */
    for (n=channel->banlist.used; n>0; ) {
	argh = channel->banlist.list[--n]->ban;
	DelChannelBan(1, &argh, channel, self->numeric, 0);
    }

    for (n=0; n<dcf_used; n++) {
        dcf_list[n](channel);
    }

    modeList_clean(&channel->members);
    banList_clean(&channel->banlist);
    free(channel);
}

struct modeNode *
AddChannelUser(userNode *user, chanNode* channel, int is_burst)
{
	modeNode* mNode;
	unsigned int n;

	mNode = GetUserMode(channel, user);
	if (mNode) return mNode; /* already in channel */

	mNode = malloc(sizeof(modeNode));

	/* set up modeNode */
	mNode->channel = channel;
	mNode->user = user;
	mNode->modes = 0;

	/* add modeNode to channel and to user */
	modeList_append(&channel->members, mNode);
	modeList_append(&user->channels, mNode);
	
	if (IsLocal(user->numeric)) {
	    irc_join(user, channel);
	}

        n = 0;
        while (n<jf_used) {
	    jf_list[n++](user, channel, is_burst);
	}

	return mNode;
}

static part_func_t *pf_list;
static unsigned int pf_size = 0, pf_used = 0;

void
reg_part_func(part_func_t handler)
{
    if (pf_used == pf_size) {
	if (pf_size) {
	    pf_size <<= 1;
	    pf_list = realloc(pf_list, pf_size*sizeof(part_func_t));
	} else {
	    pf_size = 8;
	    pf_list = malloc(pf_size*sizeof(part_func_t));
	}
    }
    pf_list[pf_used++] = handler;
}

void
unreg_part_func(part_func_t handler)
{
    unsigned int i;
    for (i=0; i<pf_used; i++) {
        if (pf_list[i] == handler) break;
    }
    if (i == pf_used) return;
    memmove(pf_list+i, pf_list+i+1, (pf_used-i-1)*sizeof(pf_list[0]));
    pf_used--;
}

void DelChannelUser(userNode* user, chanNode* channel, const char *reason, int deleting)
{
    struct modeNode* mNode;
    unsigned int n;

    if (reason) irc_part(user, channel, reason);

    mNode = GetUserMode(channel, user);

    /* Sometimes we get a PART when the user has been KICKed.
     * In this case, we get no usermode, and should not try to free it.
     */
    if (!mNode) return;

    /* remove modeNode from channel and user */
    modeList_remove(&channel->members, mNode);
    modeList_remove(&user->channels, mNode);
    free(mNode);

    for (n=0; n<pf_used; n++) pf_list[n](user, channel);

    if (!deleting && channel->members.used == 0) DelChannel(channel);
}

void
KickChannelUser(userNode* target, chanNode* channel, userNode *kicker, const char *why)
{
    char *numeric;
    if (!target || !channel || IsService(target) || !GetUserMode(channel, target)) return;

    numeric = GetUserMode(channel, kicker) ? kicker->numeric : self->numeric;

    /* don't remove them from the channel, since the server will send a PART */
    irc_kick(numeric, target, channel, why);
}

static kick_func_t *kf_list;
static unsigned int kf_size = 0, kf_used = 0;

void
reg_kick_func(kick_func_t handler)
{
    if (kf_used == kf_size) {
	if (kf_size) {
	    kf_size <<= 1;
	    kf_list = realloc(kf_list, kf_size*sizeof(kick_func_t));
	} else {
	    kf_size = 8;
	    kf_list = malloc(kf_size*sizeof(kick_func_t));
	}
    }
    kf_list[kf_used++] = handler;
}

void
ChannelUserKicked(userNode* kicker, userNode* victim, chanNode* channel)
{
    unsigned int n;

    if (!kicker || !victim || !channel || IsService(victim) || !GetUserMode(channel, victim)) return;

    for (n=0; n<kf_used; n++) {
	kf_list[n](kicker, victim, channel);
    }

    DelChannelUser(victim, channel, 0, 0);
}

static void
chan_apply_bans(const char *who, struct chanNode *channel, char *what, unsigned int count, char **list)
{
    char buffer[400];
    unsigned int start, i, here_len, queued;

    queued = 0;
    start = sizeof(buffer)-1;
    buffer[start] = 0;
    for (i=0; i<count; i++) {
	here_len = strlen(list[i]);
	if ((start < queued+here_len+3) || (queued >= MAXMODEPARAMS)) {
	    while (queued) {
		buffer[--start] = what[1];
		queued--;
	    }
	    buffer[--start] = what[0];
	    irc_mode(who, channel, buffer+start);
	    start = sizeof(buffer)-1;
	}
	start -= here_len;
	memcpy(buffer+start, list[i], here_len); /* omit terminator */
	buffer[--start] = ' ';
	queued++;
    }
    if (queued) {
	while (queued) {
	    buffer[--start] = what[1];
	    queued--;
	}
	buffer[--start] = what[0];
	irc_mode(who, channel, buffer+start);
	start = sizeof(buffer)-1;
    }
}

static void
chan_apply_mode(const char *who, struct chanNode *channel, char *what, unsigned int count, userNode **list)
{
    char buffer[400];
    unsigned int start, i, queued;
    queued = 0;
    start = sizeof(buffer)-1;
    buffer[start] = 0;

    for (i=0; i<count; i++) {
	if ((start < queued+3+COMBO_NUMERIC_LEN) || (queued >= MAXMODEPARAMS)) {
	    while (queued) {
		buffer[--start] = what[1];
		queued--;
	    }
	    buffer[--start] = what[0];
	    irc_mode(who, channel, buffer+start);
	    start = sizeof(buffer)-1;
	}
	start -= COMBO_NUMERIC_LEN;
	memcpy(buffer+start, list[i]->numeric, COMBO_NUMERIC_LEN);
	buffer[--start] = ' ';
	queued++;
    }
    if (queued) {
	while (queued) {
	    buffer[--start] = what[1];
	    queued--;
	}
	buffer[--start] = what[0];
	irc_mode(who, channel, buffer+start);
	start = sizeof(buffer)-1;
    }
}

void AddChannelBan(unsigned int num, char **banset, chanNode* channel, const char *who, time_t set, int announce)
{
    unsigned int n, i;
    struct banNode *bNode;
    const char *who_name = "<unknown>";
    struct string_list *remove = NULL;
    struct banNode **bans;
    char *ban;

    /* See if we can figure out who to credit for the ban. */
    if (who) {
        if (who[0] == ':') {
            who_name = who + 1;
        } else {
            struct userNode *un;
            if ((un = GetUserN(who))) who_name = un->nick;
        }
    }

    for (n=0; n<num; n++) {
	if(channel->banlist.used >= MAXBANS)
	{
	    break;
	}

	/* We check for bans that are a superset of an existing ban
	   here and silently remove the old ban to be compatible with
	   ircu; however, we _do_not_ check for bans that are a subset
	   of an existing ban here, instead, that is checked in the
	   user interface code (because it blocks the ban).

	   We can skip the checking during bursts, because bans come
	   from the uplink's internal banlists during a burst. */
	if(!burst) {
	    bans = channel->banlist.list;
	    for(i = 0; i < channel->banlist.used; i++) {
		if(match_ircglobs(banset[n], bans[i]->ban)) {
		    if(!remove) {
			remove = alloc_string_list(2);
		    }

		    ban = strdup(bans[i]->ban);
		    string_list_append(remove, ban);
		}
	    }
	}

	bNode = malloc(sizeof(struct banNode));
	
	safestrncpy(bNode->ban, banset[n], sizeof(bNode->ban));
	safestrncpy(bNode->who, who_name, sizeof(bNode->who));
	bNode->set = set;
	banList_append(&channel->banlist, bNode);
    }

    /* If we have to announce the new ban, announce the
       removal of the old; if we aren't announcing the new
       ban (we received the ban from our uplink), do it
       silently. */
    if(remove) {
	DelChannelBan(remove->used, remove->list, channel, who, announce);
	free_string_list(remove);
    }
    if (announce) chan_apply_bans(who, channel, "+b", n, banset);
}

void DelChannelBan(unsigned int num, char **banset, chanNode* channel, const char *who, int announce)
{
    unsigned int n, m;
    struct banNode *bNode = NULL; /* bogus, gcc-2.97 warns spuriously */

    for (n=0; n<num; n++) {
	for (m=0; m<channel->banlist.used; m++) {
	    bNode = channel->banlist.list[m];
	    if (!strcmp(bNode->ban, banset[n])) break;
	}
	if (m < channel->banlist.used) {
	    banList_remove(&channel->banlist, bNode);
	    free(bNode);
	}
    }

    if (announce) chan_apply_bans(who, channel, "-b", num, banset);
}

void AddChannelOp(unsigned int num, userNode **newops, chanNode *channel, const char *who, int announce)
{
    unsigned int n;
    struct modeNode *mn;

    for (n=0; n<num; n++) {
	if (!(mn = GetUserMode(channel, newops[n]))) continue;
	mn->modes |= MODE_CHANOP;
    }
    if (announce) chan_apply_mode(who, channel, "+o", num, newops);
}

void DelChannelOp(unsigned int num, userNode **exops, chanNode *channel, const char *who, int announce)
{
    unsigned int n;
    struct modeNode *mn;

    for (n=0; n<num; n++) {
	if (!(mn = GetUserMode(channel, exops[n]))) continue;
	mn->modes &= ~MODE_CHANOP;
    }
    if (announce) chan_apply_mode(who, channel, "-o", num, exops);
}

void AddChannelVoice(unsigned int num, userNode **newv, chanNode *channel, const char *who, int announce)
{
    unsigned int n;
    struct modeNode *mn;

    for (n=0; n<num; n++) {
	if (!(mn = GetUserMode(channel, newv[n]))) continue;
	mn->modes |= MODE_VOICE;
    }
    if (announce) chan_apply_mode(who, channel, "+v", num, newv);
}

void DelChannelVoice(unsigned int num, userNode **exv, chanNode *channel, const char *who, int announce)
{
    unsigned int n;
    struct modeNode *mn;

    for (n=0; n<num; n++) {
	if (!(mn = GetUserMode(channel, exv[n]))) continue;
	mn->modes &= ~MODE_VOICE;
    }
    if (announce) chan_apply_mode(who, channel, "-v", num, exv);
}

static topic_func_t *tf_list;
static unsigned int tf_size = 0, tf_used = 0;

void
reg_topic_func(topic_func_t handler)
{
    if (tf_used == tf_size) {
	if (tf_size) {
	    tf_size <<= 1;
	    tf_list = realloc(tf_list, tf_size*sizeof(topic_func_t));
	} else {
	    tf_size = 8;
	    tf_list = malloc(tf_size*sizeof(topic_func_t));
	}
    }
    tf_list[tf_used++] = handler;
}

void SetChannelTopic(chanNode *channel, const char *nick, const char *topic, int announce)
{
    unsigned int n;

    if (topic) {
        safestrncpy(channel->topic, topic, sizeof(channel->topic));
    }
    if (nick) {
        safestrncpy(channel->topic_nick, nick, sizeof(channel->topic_nick));
    }
    channel->topic_time = now;

    if (!announce) {
	/* We don't really care if a local user messes with the topic. */
	for (n=0; n<tf_used; n++) {
	    tf_list[n](nick, channel);
	}
    } else {
	irc_topic(nick, channel, topic);
    }
}

chanNode* GetChannel(const char *name)
{
    return dict_find(channels, name, NULL);
}

modeNode* GetUserMode(chanNode* channel, userNode* user)
{
    unsigned int n;
    struct modeNode *mn = NULL;
    if (channel->members.used < user->channels.used) {
	for (n=0; n<channel->members.used; n++) {
	    if (user == channel->members.list[n]->user) {
		mn = channel->members.list[n];
		break;
	    }
	}
    } else {
	for (n=0; n<user->channels.used; n++) {
	    if (channel == user->channels.list[n]->channel) {
		mn = user->channels.list[n];
		break;
	    }
	}
    }
    return mn;
}

int IsLocal(const char *numeric)
{
    return (self && numeric && !strncmp(numeric, self->numeric, SERVER_NUMERIC_LEN));
}

DEFINE_LIST(userList, struct userNode*);
DEFINE_LIST(modeList, struct modeNode*);
DEFINE_LIST(banList, struct banNode*);
DEFINE_LIST(channelList, struct chanNode*);
