/* proto.c - IRC protocol output
 * Copyright 2000-2001 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: proto.c,v 1.68.2.14 2001/11/06 18:01:18 entrope Exp $
 */

#include "config.h"

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "common.h"
#include "commands.h"
#include "conf.h"
#include "hash.h"
#include "log.h"
#include "parse.h"
#include "timeq.h"

int socket_fd = -1;
unsigned int lines_processed;
FILE *replay_file;
char replay_line[MAXLEN+80];

extern struct cManagerNode cManager;
extern unsigned long burst_length;

int
create_socket_client(struct uplinkNode *target)
{
    int		socket_desc;
    int		port = target->port;
    const	char		*addr = target->host;
    struct	sockaddr_in	sain;
    struct	hostent		*host_info;

    if (replay_file) {
        return feof(replay_file) ? 0 : 1;
    }

    log(MAIN_LOG, LOG_INFO, "Connecting to %s:%i...\n", addr, port);

    if (socket_fd != -1) {
	/* Leave the existing socket open, say we failed. */
	log(MAIN_LOG, LOG_ERROR, "Refusing to create second connection to %s:%d.\n", addr, port);
	return 0;
    }

    if ((host_info = gethostbyname(addr)) == NULL) {
	log(MAIN_LOG, LOG_ERROR, "Resolve failed: %s\n", strerror(errno));
	target->state = DISCONNECTED;
	target->tries++;
	return 0;
    }

    memcpy((char*)&(sain.sin_addr), host_info->h_addr, host_info->h_length);
    sain.sin_family = host_info->h_addrtype;
    sain.sin_port = htons((u_short) port);

    if ((socket_desc = socket(host_info->h_addrtype, SOCK_STREAM, 0)) < 0) {
	log(MAIN_LOG, LOG_ERROR, "Socket create failed: %s\n", strerror(errno));
	log(MAIN_LOG, LOG_ERROR, MSG_FATAL_ERROR);
	exit(1);
    }
    if (cManager.uplink->bind_addr) {
	if (bind(socket_desc, (struct sockaddr*) cManager.uplink->bind_addr, sizeof(struct sockaddr)) < 0) {
	    log(MAIN_LOG, LOG_ERROR, "Error: bind() in socket setup returned errno %d (%s)\n", errno, strerror(errno));
            log(MAIN_LOG, LOG_ERROR, MSG_FATAL_ERROR);
	    exit(1);
	}
    }
    if (connect(socket_desc, (struct sockaddr *)&sain, sizeof(sain)) < 0) {
	log(MAIN_LOG, LOG_ERROR, "Connect failed: %s\n", strerror(errno));
	switch(errno)
	{
	case ECONNREFUSED:
	case ETIMEDOUT:
	case ENETUNREACH:
	case EHOSTUNREACH:
	case EAGAIN:
	    target->state = DISCONNECTED;
	    target->tries++;
	    return 0;
	default:
	    log(MAIN_LOG, LOG_ERROR, MSG_FATAL_ERROR);
	    exit(1);
	}
    }

    log(MAIN_LOG, LOG_INFO, "Connection to server established.\n");
    socket_fd = socket_desc;
    cManager.uplink = target;

    target->state = AUTHENTICATING;
    target->tries = 0;

    return 1;
}

void
replay_read_line(void)
{
    struct tm timestamp;
    time_t new_time;

    if (replay_line[0]) return;
  read_line:
    if (!fgets(replay_line, sizeof(replay_line), replay_file)) {
        if (feof(replay_file)) {
            extern int quit_services;
            quit_services = 1;
            memset(replay_line, 0, sizeof(replay_line));
            return;
        }
    }
    if ((replay_line[0] != '[')
        || (replay_line[3] != ':')
        || (replay_line[6] != ':')
        || (replay_line[9] != ' ')
        || (replay_line[12] != '/')
        || (replay_line[15] != '/')
        || (replay_line[20] != ']')
        || (replay_line[21] != ' ')) {
        log(MAIN_LOG, LOG_ERROR, "Unrecognized timestamp in replay file: %s\n", replay_line);
        goto read_line;
    }
    timestamp.tm_hour = strtoul(replay_line+1, NULL, 10);
    timestamp.tm_min = strtoul(replay_line+4, NULL, 10);
    timestamp.tm_sec = strtoul(replay_line+7, NULL, 10);
    timestamp.tm_mon = strtoul(replay_line+10, NULL, 10) - 1;
    timestamp.tm_mday = strtoul(replay_line+13, NULL, 10);
    timestamp.tm_year = strtoul(replay_line+16, NULL, 10) - 1900;
    timestamp.tm_isdst = 0;
    new_time = mktime(&timestamp);
    if (new_time == -1) {
        log(MAIN_LOG, LOG_ERROR, "Unable to parse time struct tm_sec=%d tm_min=%d tm_hour=%d tm_mday=%d tm_mon=%d tm_year=%d\n", timestamp.tm_sec, timestamp.tm_min, timestamp.tm_hour, timestamp.tm_mday, timestamp.tm_mon, timestamp.tm_year);
    } else {
        now = new_time;
    }

    if (strncmp(replay_line+22, "(info) ", 7)) goto read_line;
    return;
}

static int
replay_read(void)
{
    size_t len;
    unsigned char read_line[MAXLEN];
    while (1) {
        replay_read_line();
        /* if it's a sent line, break out to handle it */
        if (!strncmp(replay_line+29, "   ", 3)) break;
        if (!strncmp(replay_line+29, "W: ", 3)) {
            log(MAIN_LOG, LOG_ERROR, "Expected response from services: %s", replay_line+32);
            replay_line[0] = 0;
        } else {
            return 0;
        }
    }
    log(DEBUG_LOG, LOG_INFO, "   %s", replay_line+32);
    safestrncpy(read_line, replay_line+32, sizeof(read_line));
    len = strlen(read_line);
    if (read_line[len-1] == '\n') read_line[--len] = 0;
    replay_line[0] = 0;
    parse_line(read_line, 0);
    lines_processed++;
    return 1;
}

int
read_from_socket(void)
{
    static char in[MAXLEN];
    static int used = 0;
    static char *buffer;
    int i, recv_len;

    if (replay_file) return replay_read();

    recv_len = read(socket_fd, in+used, sizeof(in)-used);
    if (recv_len < 0) {
	/* if it would have blocked, pretend we succeeded */
	if ((errno == EWOULDBLOCK) || (errno == EAGAIN)) return 1;
	return 0;
    }
    if (!recv_len) {
	/* end of file encountered */
	cManager.uplink->state = DISCONNECTED;
	return 0;
    }
    buffer = in;
    for (i=0; i<recv_len+used; i++) {
	if (in[i] == '\r') in[i] = 0;
	if (in[i] == '\n') {
	    in[i] = 0;
	    log(DEBUG_LOG, LOG_INFO, "   %s\n", buffer);
	    parse_line(buffer, 0);
	    buffer = in + i + 1;
	    lines_processed++;
	}
    }
    used = (used + recv_len) - (buffer - in);
    memmove(in, buffer, used);
    return 1;
}

static void
replay_write(unsigned char *text)
{
    replay_read_line();
    if (strncmp(replay_line+29, "W: ", 3)) {
        log(MAIN_LOG, LOG_ERROR, "Unexpected output during replay: %s", text);
        return;
    } else {
        if (strcmp(replay_line+32, text)) {
            log(MAIN_LOG, LOG_ERROR, "Incorrect output during replay:\nReceived: %sExpected: %s", text, replay_line+32);
        } else {
            log(DEBUG_LOG, LOG_INFO, "W: %s", text);
        }
        replay_line[0] = 0;
    }
}


#ifdef __GNUC__
void putsock(const char *text, ...) __attribute__((format (printf, 1, 2)));
#endif

void
putsock(const char *text, ...)
{
    va_list	arg_list;
    unsigned char buffer[MAXLEN];
    int pos;

    if (!cManager.uplink || cManager.uplink->state == DISCONNECTED) return;
    buffer[0] = '\0';
    va_start(arg_list, text);
    pos = vsnprintf(buffer, MAXLEN - 2, text, arg_list);
    if (pos < 0 || pos > (MAXLEN - 2)) pos = MAXLEN - 2;
    /* buffer[pos++] = '\r'; */
    buffer[pos++] = '\n';
    buffer[pos] = 0;
    va_end(arg_list);
    if (!replay_file) {
        log(DEBUG_LOG, LOG_INFO, "W: %s", buffer);
        if (send(socket_fd, buffer, pos, 0) < 0) {
            log(MAIN_LOG, LOG_ERROR, "Write failed: %s\n", strerror(errno));
            if (errno != EAGAIN) close_socket();
        }
    } else {
        replay_write(buffer);
    }
}

void
close_socket(void)
{
    close(socket_fd);
    socket_fd = -1;
    cManager.uplink->state = DISCONNECTED;
}

void
irc_server(struct server *srv)
{
    putsock("%s %s %s %d %li %li P10 %s]] :%s",
	    self->numeric, CMD_SERVER, srv->name, srv->hops, srv->boot, srv->link, srv->numeric, srv->description);
}

void
irc_jserver(struct server *srv)
{
    putsock("%s %s %d %li %li J10 %s]] :%s",
	    CMD_SERVER, srv->name, srv->hops + 1, srv->boot, srv->link, srv->numeric, srv->description);
}

void
irc_uptime(struct userNode *user)
{
    char uptime[INTERVALLEN];
    extern time_t boot_time;

    intervalString(uptime, now-boot_time);
    putsock(":%s 242 %s :Server Up %s", self->name, user->nick, uptime);
    putsock(":%s 219 %s u :End of /STATS report", self->name, user->nick);
}

void
irc_whois(struct userNode *from, struct userNode *who)
{
    putsock(":%s 311 %s %s %s %s * :%s", self->name, from->nick, who->nick, who->username, who->hostname, who->info);
    putsock(":%s 312 %s %s %s :%s", self->name, from->nick, who->nick, self->name, self->description);
    if(IsOper(who))
     putsock(":%s 313 %s %s :is a megalomaniacal power hungry tyrant", self->name, from->nick, who->nick);
    putsock(":%s 318 %s %s :End of /WHOIS list.", self->name, from->nick, who->nick);
}

void
irc_user(struct userNode *user)
{
    char b64ip[7];
    if (!user) return;
    inttobase64(b64ip, user->ip, 6);
    if (user->modes) {
	int modelen = 0;
	char modes[9];
	if (IsOper(user)) modes[modelen++] = 'o';
	if (IsInvisible(user)) modes[modelen++] = 'i';
	if (IsWallOp(user)) modes[modelen++] = 'w';
	if (IsService(user)) modes[modelen++] = 'k';
	if (IsServNotice(user)) modes[modelen++] = 's';
	if (IsDeaf(user)) modes[modelen++] = 'd';
	if (IsGlobal(user)) modes[modelen++] = 'g';
	if (IsHelperIrcu(user)) modes[modelen++] = 'h';
	modes[modelen] = 0;

        /* we don't need to put the + in modes because it's in the format string. */
	putsock("%s %s %s %d %li %s %s +%s %s %s :%s",
		user->uplink->numeric, CMD_NICK, user->nick, user->uplink->hops+1, user->timestamp, user->username, user->hostname, modes, b64ip, user->numeric, user->info);
    } else {
	putsock("%s %s %s %d %li %s %s %s %s :%s",
		user->uplink->numeric, CMD_NICK, user->nick, user->uplink->hops+1, user->timestamp, user->username, user->hostname, b64ip, user->numeric, user->info);
    }
}

void
irc_nick(struct userNode *user)
{
    if (!user) return;
    putsock("%s %s %s "FMT_TIME_T"", user->numeric, CMD_NICK, user->nick, now);
}

void
irc_fetchtopic(const char *from, const char *to)
{
    if (!from || !to) return;
    putsock("%s TOPIC %s", from, to);
}

void
irc_squit(struct server *srv, const char *message)
{
    /* If we're squitting our uplink, services is leaving the network. */
    if(srv == self->uplink && cManager.uplink->state == CONNECTED)
    {
	extern unsigned int highest_numeric;
	unsigned int i;

	/* Quit all services. */
	for (i = 0; i < highest_numeric; i++)
	{
	    if(!self->users[i])
	    {
		continue;
	    }

	    irc_quit(self->users[i], message);
	}
    }

    putsock("%s %s %s %d :%s", self->numeric, CMD_SQUIT, srv->name, 0, message);

    if(srv == self->uplink)
    {
	/* Force a reconnect to the currently selected server. */
	cManager.uplink->tries = 0;
        log(MAIN_LOG, LOG_INFO, "Squitting from uplink: %s\n", message);
	close_socket();
    }
}

void
irc_notice(const char *from, const char *to, const char *message)
{
    putsock("%s %s %s :%s", from, CMD_NOTICE, to, message);
}

void
irc_privmsg(const char *from, const char *to, const char *message)
{
    putsock("%s %s %s :%s", from, CMD_PRIVMSG, to, message);
}

void
irc_eob(void)
{
    putsock("%s %s", self->numeric, CMD_EOB);
}

void
irc_eob_ack(void)
{
    putsock("%s %s", self->numeric, CMD_EOB_ACK);
}

void
irc_ping(const char *something)
{
    putsock("%s %s :%s", self->numeric, CMD_PING, something);
}

void
irc_pong(const char *data)
{
    putsock("%s %s %s :%s", self->numeric, CMD_PONG, self->name, data);
}

void
irc_pass(const char *passwd)
{
    putsock("PASS :%s", passwd);
}

void
irc_introduce(const char *passwd)
{
    void timed_send_ping(void *data);
    extern int burst, ping_freq;
    extern time_t burst_begin;

    irc_pass(passwd);
    irc_jserver(self);
    burst = 1;
    burst_length = 0;
    burst_begin = now;
    timeq_add(now + ping_freq, timed_send_ping, 0);
}

void
irc_gline(const char *mask, unsigned long duration, const char *message)
{
    putsock("%s %s * +%s %ld :%s", self->numeric, CMD_GLINE, mask, duration, message);
}

void
irc_ungline(const char *mask)
{
    putsock("%s %s * -%s", self->numeric, CMD_GLINE, mask);
}

void
irc_jupe(const char *server, unsigned long duration, const char *reason)
{
    putsock("%s %s * +%s %ld "FMT_TIME_T" :%s", self->numeric, CMD_JUPE, server, duration, now, reason);
}

void
irc_unjupe(const char *server)
{
    putsock("%s %s * -%s 0 "FMT_TIME_T" :Unjuping.", self->numeric, CMD_JUPE, server, now);
}

int
irc_make_chanmode(struct chanNode *chan, char *out)
{
    int pos = 0;

    if (!chan->modes)
    {
	out[0] = 0;
	return 0;
    }

    out[pos++] = '+';
#define do_chan_mode(MODE,CHAR) if (chan->modes & MODE) out[pos++] = CHAR
    do_chan_mode(MODE_PRIVATE, 'p');
    do_chan_mode(MODE_SECRET, 's');
    do_chan_mode(MODE_MODERATED, 'm');
    do_chan_mode(MODE_TOPICLIMIT, 't');
    do_chan_mode(MODE_INVITEONLY, 'i');
    do_chan_mode(MODE_NOPRIVMSGS, 'n');
    do_chan_mode(MODE_KEY, 'k');
    do_chan_mode(MODE_LIMIT, 'l');
#undef do_chan_mode
    if (chan->modes & MODE_KEY) pos += sprintf(out+pos, " %s", chan->key);
    if (chan->modes & MODE_LIMIT) pos += sprintf(out+pos, " %d", chan->limit);
    out[pos++] = 0;
    return pos;
}

void
irc_burst(struct chanNode *chan)
{
    char burst_line[512];
    int pos, base_len, len;
    struct modeNode *mn;
    struct banNode *bn;
    long last_mode=-1;
    unsigned int n;

    base_len = sprintf(burst_line, "%s %s %s " FMT_TIME_T " ", self->numeric, CMD_BURST, chan->name, chan->timestamp);
    len = irc_make_chanmode(chan, burst_line+base_len);
    pos = base_len + len - 1;
    burst_line[pos++] = ' ';

    /* dump the users */
    for (n=0; n<chan->members.used; n++) {
	mn = chan->members.list[n];
	if (pos > 500) {
	    burst_line[pos-1] = 0; /* -1 to back up over the space or comma */
	    putsock("%s", burst_line);
	    pos = base_len;
	    last_mode = -1;
	}
	memcpy(burst_line+pos, mn->user->numeric, 5);
	pos += strlen(mn->user->numeric);
	if (mn->modes && (mn->modes != last_mode)) {
	    last_mode = mn->modes;
	    burst_line[pos++] = ':';
	    if (last_mode & MODE_CHANOP) burst_line[pos++] = 'o';
	    if (last_mode & MODE_VOICE) burst_line[pos++] = 'v';
	}
	if ((n+1)<chan->members.used) burst_line[pos++] = ',';
    }
    if (chan->banlist.used) {
	/* dump the bans */
	if (pos+2+strlen(chan->banlist.list[0]->ban) > 505) {
	    burst_line[pos-1] = 0;
	    putsock("%s", burst_line);
	    pos = base_len;
	}
	burst_line[pos++] = ':';
	burst_line[pos++] = '%';
	base_len = pos;
	for (n=0; n<chan->banlist.used; n++) {
	    bn = chan->banlist.list[n];
	    len = strlen(bn->ban);
	    if (pos+len+1 > 510) {
		burst_line[pos-1] = 0; /* -1 to back up over the space or comma */
		putsock("%s", burst_line);
		pos = base_len;
	    }
	    memcpy(burst_line+pos, bn->ban, len);
	    pos += len;
	    burst_line[pos++] = ' ';
	}
    }
    /* print the last line */
    burst_line[pos] = 0;
    putsock("%s", burst_line);
}

void
irc_quit(struct userNode *user, const char *message)
{
    putsock("%s %s :%s", user->numeric, CMD_QUIT, message);
}

void
irc_error(const char *to, const char *message)
{
    if (to) {
	putsock("%s %s :%s", to, CMD_ERROR, message);
    } else {
	putsock(":%s %s :%s", self->name, CMD_ERROR, message);
    }
}

void
irc_kill(struct userNode *from, struct userNode *target, const char *message)
{
    if (from) {
	putsock("%s %s %s :%s!%s (%s)", from->numeric, CMD_KILL, target->numeric, self->name, from->nick, message);
    } else {
	putsock("%s %s %s :%s (%s)", self->numeric, CMD_KILL, target->numeric, self->name, message);
    }
}

void
irc_mode(const char *from, struct chanNode *target, const char *modes)
{
    putsock("%s %s %s %s "FMT_TIME_T, from, CMD_MODE, target->name, modes, target->timestamp);
}

void
irc_invite(const char *from, struct userNode *who, chanNode *to)
{
    putsock("%s %s %s %s", from, CMD_INVITE, who->nick, to->name);
}

void
irc_join(struct userNode *who, struct chanNode *what)
{
    if (what->members.used == 1) {
	putsock("%s %s %s %lu", who->numeric, CMD_CREATE, what->name, what->timestamp);
    } else {
	putsock("%s %s %s", who->numeric, CMD_JOIN, what->name);
    }
}

void
irc_kick(char *numeric, struct userNode *who, struct chanNode *from, const char *msg)
{
    putsock("%s %s %s %s :%s", numeric, CMD_KICK, from->name, who->numeric, msg);
}

void
irc_raw(const char *what)
{
    putsock("%s", what);
}

void
irc_stats(struct userNode *from, struct server *target, char type)
{
    putsock("%s STATS %c :%s", from->numeric, type, target->numeric);
}

void
irc_part(struct userNode *who, struct chanNode *what, const char *reason)
{
    putsock("%s %s %s :%s", who->numeric, CMD_PART, what->name, reason);
}

void
irc_topic(const char *who, struct chanNode *what, const char *topic)
{
    putsock(":%s %s %s :%s", who, CMD_TOPIC, what->name, topic);
}
